package cc.smtweb.system.bpm.engine.process.impl.dao;

import cc.smtweb.framework.core.SwMap;
import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.db.cache.ModelTableCache;
import cc.smtweb.framework.core.db.sqlbuilder.InsertSqlBuilder;
import cc.smtweb.framework.core.db.sqlbuilder.SqlBuilder;
import cc.smtweb.framework.core.db.sqlbuilder.UpdateSqlBuilder;
import cc.smtweb.framework.core.db.vo.ModelTable;
import cc.smtweb.framework.core.util.DateUtil;
import cc.smtweb.system.bpm.core.exception.BpmLockRevException;
import cc.smtweb.system.bpm.core.exception.BpmObjectNotFoundException;
import cc.smtweb.system.bpm.engine.process.impl.entity.ProcessLogPO;
import cc.smtweb.system.bpm.engine.process.impl.entity.ProcessPO;
import cc.smtweb.system.bpm.engine.process.impl.entity.TaskPO;
import cc.smtweb.system.bpm.engine.process.impl.entity.VariablePO;
import cc.smtweb.system.bpm.engine.process.runtime.ProcessInstance;
import cc.smtweb.system.bpm.engine.process.task.Task;
import cc.smtweb.system.bpm.engine.process.vo.Flow;
import cc.smtweb.system.bpm.engine.process.vo.FlowNode;
import cc.smtweb.system.bpm.spring.cache.BpmFlowCache;
import org.apache.commons.lang3.StringUtils;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;

public class TaskDao {
  private DbEngine dbEngine;

  public TaskDao(DbEngine dbEngine) {
    this.dbEngine = dbEngine;
  }

  public void saveProcess(ProcessInstance processInstance, Timestamp now) {
    String route = processInstance.postRoute();
    long lastTime = DateUtil.nowDateTimeLong();

    if (processInstance.getCreateAt() == 0L) {
      Flow flow = processInstance.getFlow();

      InsertSqlBuilder sqlBuilder = SqlBuilder.createInsert("sw_bpm.bpm_process");
      sqlBuilder.add("process_id", processInstance.getId())
              .add("process_parent_id", processInstance.getParentId())
              .add("process_root_id", processInstance.getRootId())
              .add("process_flow_id", flow.getId())
              .add("process_business_key", processInstance.getBusinessKey())
              .add("process_name", processInstance.getName())
              .add("process_node_key", processInstance.getCurNode().getId())
              .add("process_priority", processInstance.getPriority())
              .add("process_active", 1)
              .add("process_route", route)
              .add("process_is_concurrent", processInstance.isConcurrent() ? 1 : 0)
              .add("process_create_time", lastTime)
              .add("process_last_time", lastTime);
      sqlBuilder.update(dbEngine);

      processInstance.setCreateAt(DateUtil.nowDateTimeLong());
    } else {
      UpdateSqlBuilder sqlBuilder = SqlBuilder.createUpdate("sw_bpm.bpm_process");
      sqlBuilder.add("process_business_key", processInstance.getBusinessKey())
              .add("process_node_key", processInstance.getCurNode().getId())
              .add("process_priority", processInstance.getPriority())
              .add("process_route", route)
              .add("process_lock_rev", processInstance.getLockRev() + 1)
              .add("process_last_time", lastTime)
              .addWhere("process_id", processInstance.getId())
              .addWhere("process_parent_id", processInstance.getLockRev());
      int ret = sqlBuilder.update(dbEngine);

      if (ret == 0) {
        throw new BpmLockRevException();
      }

      processInstance.incLockRev();
    }

    processInstance.setUpdateAt(DateUtil.nowDateTimeLong());
  }

  public void saveTask(Task task, Timestamp now) {
    if (task.getCreateAt() == 0L) {
      ProcessInstance processInstance = task.getProcessInstance();

      InsertSqlBuilder sqlBuilder = SqlBuilder.createInsert("sw_bpm.bpm_task");
      sqlBuilder.add("task_id", task.getId())
              .add("task_parent_id", task.getParentId())
              .add("task_process_id", processInstance.getId())
              .add("task_process_root_id", processInstance.getRootId())
              .add("task_ui_id", task.getUiId())
              .add("task_assignee_uid", task.getAssigneeUserId())
              .add("task_name", task.getName())
              .add("task_priority", task.getPriority());
      sqlBuilder.update(dbEngine);
    } else {
      UpdateSqlBuilder sqlBuilder = SqlBuilder.createUpdate("sw_bpm.bpm_task");
      sqlBuilder.add("task_assignee_uid", task.getAssigneeUserId())
              .add("task_name", task.getName())
              .add("task_priority", task.getPriority())
              .add("task_lock_rev", task.getLockRev() + 1)
              .add("task_last_time", DateUtil.nowDateTimeLong())
              .addWhere("task_id", task.getId())
              .addWhere("task_lock_rev", task.getLockRev());
      int ret = sqlBuilder.update(dbEngine);

      if (ret == 0) {
        throw new BpmLockRevException();
      }

      task.incLockRev();
    }

    task.setUpdateAt(DateUtil.nowDateTimeLong());
  }

  public void suspensionTask(Task task) {
    // 修改挂起状态
    dbEngine.update("update sw_bpm.bpm_task set task_is_suspension=1 where task_id=?", task.getId());
  }

  public void deleteProcess(ProcessInstance processInstance) {
    String route = processInstance.postRoute();
    // 添加日志
    InsertSqlBuilder sqlBuilder = SqlBuilder.createInsert("sw_bpm.bpm_process_log");
    sqlBuilder.add("pl_id", processInstance.getId())
            .add("pl_parent_id", processInstance.getParentId())
            .add("pl_root_id", processInstance.getRootId())
            .add("pl_flow_id", processInstance.getFlow().getId())
            .add("pl_business_key", processInstance.getBusinessKey())
            .add("pl_name", processInstance.getName())
            .add("pl_delete_reason", 1)
            .add("pl_priority", processInstance.getPriority())
            .add("pl_is_concurrent", processInstance.isConcurrent() ? 1 : 0)
            .add("pl_route", route)
            .add("pl_begin_time", processInstance.getCreateAt())
            .add("pl_end_time", processInstance.getUpdateAt())
            .add("pl_duration", tsDur(processInstance.getCreateAt(), processInstance.getUpdateAt()));
    sqlBuilder.update(dbEngine);

    // 删除运行时数据
    dbEngine.update("delete from sw_bpm.bpm_process where process_id=?", processInstance.getId());
  }

  public void deleteTask(Task task) {
    if (task.getCreateAt() > 0L) {
      // 添加日志
      ProcessInstance process = task.getProcessInstance();

      InsertSqlBuilder sqlBuilder = SqlBuilder.createInsert("sw_bpm.bpm_task_log");
      sqlBuilder.add("tl_id", task.getId())
              .add("tl_parent_id", task.getParentId())
              .add("tl_process_id", process.getId())
              .add("tl_process_root_id", process.getRootId())
              .add("tl_ui_id", task.getUiId())
              .add("tl_assignee_uid", task.getAssigneeUserId())
              .add("tl_owner_uid", task.getOwnerUserId())
              .add("tl_delegation_status", 0)
              .add("tl_name", task.getName())
              .add("tl_delete_reason", 1)
              .add("tl_node_key", task.getNodeKey())
              .add("tl_priority", task.getPriority())
              .add("tl_is_suspension", task.getIsSuspension())
              .add("tl_begin_time", task.getCreateAt())
              .add("tl_end_time", task.getUpdateAt())
              .add("tl_duration", tsDur(task.getCreateAt(), task.getUpdateAt()));
      sqlBuilder.update(dbEngine);

      // 删除数据
      dbEngine.update("delete from sw_bpm.bpm_task where task_id=?", task.getId());
      // 避免并行节点重复删除
      task.setCreateAt(0L);
    }
  }

  private long tsDur(long begin, long end) {
    long diff = (end - begin);// / 1000;

    long result = 0;
    if (diff > 0) {
      // sec
      result = diff % 60;
      diff /= 60;
      // min
      result += (diff % 60) * 100;
      diff /= 60;
      // hour
      result += (diff % 24) * 10000;
      // day
      result += diff / 24;
    }

    return result;
  }

  public Task loadTask(long taskId) {
    TaskPO taskPO = dbEngine.queryEntity("select task_id,task_parent_id,task_process_id,task_process_root_id,task_ui_id,task_assignee_uid,task_name,task_priority,task_lock_rev,task_create_time,task_last_time from sw_bpm.bpm_task where task_id=?", TaskPO.class, taskId);

    if (taskPO == null) {
      throw new BpmObjectNotFoundException("未找到任务, task_id=" + taskId);
    }

    Task task = new Task(dbEngine.nextId(), loadProcess(taskPO.getTaskProcessId()));
    task.setId(taskId);
    task.setAssigneeUserId(taskPO.getTaskAssigneeUid());
    task.setLockRev(taskPO.getTaskLockRev());
    task.setName(taskPO.getTaskName());
    task.setParentId(taskPO.getTaskParentId());
    task.setPriority(taskPO.getTaskPriority());
    task.setCreateAt(taskPO.getTaskCreateTime());
    task.setUpdateAt(taskPO.getTaskLastTime());

    return task;
  }

  public List<ProcessInstance> loadProcessList(List<Long> ids) {
    List<ProcessPO> processPos = dbEngine.query("select process_id,process_parent_id,process_root_id,process_flow_id,process_business_key,process_node_key,process_priority,process_active,process_is_concurrent,process_lock_rev,process_create_time,process_last_time,process_route from sw_bpm.bpm_process where process_id in("
                    + StringUtils.join(ids, ",") + ")", ProcessPO.class);

    if (processPos != null) {
      List<ProcessInstance> result = new ArrayList<>();
      for (ProcessPO processPo : processPos) {
        result.add(buildProcessInstance(processPo));
      }

      return result;
    }

    return null;
  }

  public ProcessInstance loadProcess(long processId) {
    ProcessPO processPo = dbEngine.queryEntity("select process_id,process_parent_id,process_root_id,process_flow_id,process_business_key,process_node_key,process_priority,process_active,process_is_concurrent,process_lock_rev,process_create_time,process_last_time,process_route from sw_bpm.bpm_process where process_id=?",
            ProcessPO.class, processId);

    if (processPo == null) {
      throw new BpmObjectNotFoundException("未找到流程实例, process_id=" + processId);
    }

    return buildProcessInstance(processPo);
  }

  private ProcessInstance buildProcessInstance(ProcessPO processPo) {
    ProcessInstance processInstance = startProcessInstanceByKey(processPo.getProcessFlowId(), processPo.getProcessNodeKey());
    processInstance.setId(processPo.getProcessId());
    processInstance.setRootId(processPo.getProcessRootId());
    processInstance.setLockRev(processPo.getProcessLockRev());
    processInstance.setBusinessKey(processPo.getProcessBusinessKey());
    processInstance.setPriority(processPo.getProcessPriority());
    processInstance.setParentId(processPo.getProcessParentId());
    processInstance.fillRoute(processPo.getProcessRoute());
    processInstance.setCreateAt(processPo.getProcessCreateAt());
    processInstance.setUpdateAt(processPo.getProcessUpdateAt());
    processInstance.setConcurrent(processPo.getProcessIsConcurrent() != 0);
    return processInstance;
  }

  public ProcessInstance loadProcessByLog(Long processId) {
    ProcessLogPO processPO = dbEngine.queryEntity("select pl_id,pl_parent_id,pl_root_id,pl_flow_id,pl_business_key,pl_node_key,pl_priority,pl_is_concurrent,pl_create_time,pl_route from sw_bpm.bpm_process_log where pl_id=?",
            ProcessLogPO.class, processId);

    if (processPO == null) {
      throw new BpmObjectNotFoundException("未找到流程实例, process_id=" + processId);
    }

    ProcessInstance processInstance = startProcessInstanceByKey(processPO.getPlFlowId(), null);
    processInstance.setId(processPO.getPlId());
    processInstance.setRootId(processPO.getPlRootId());
    processInstance.setBusinessKey(processPO.getPlBusinessKey());
    processInstance.setPriority(processPO.getPlPriority());
    processInstance.setParentId(processPO.getPlParentId());
    processInstance.fillRoute(processPO.getPlRoute());
    processInstance.setCreateAt(processPO.getPlCreateAt());
    processInstance.setUpdateAt(processPO.getPlCreateAt());
    processInstance.setConcurrent(processPO.getPlIsConcurrent() != 0);
    return processInstance;
  }


  public SwMap loadVars(ProcessInstance processInstance) {
    // 读取流程变量
    List<VariablePO> list = dbEngine.query("select var_id,var_name,var_type,var_scope,var_value from bpm_variable where (var_root_process_id=? and var_scope=1) or (var_process_id=? and var_scope=2)",
            VariablePO.class, processInstance.getRootId(), processInstance.getId());

    if (list != null) {
      SwMap vars = new SwMap(list.size());
      for (VariablePO variablePo : list) {
        String name = variablePo.getVarName();
        String value = variablePo.getVarValue();

        if (StringUtils.isNotBlank(value)) {
          switch (variablePo.getType()) {
            case VariablePO.TYPE_CHAR:
              vars.put(name, value);
              break;
            case VariablePO.TYPE_INT:
              vars.put(name, Long.parseLong(value));
              break;
            case VariablePO.TYPE_FLOAT:
              vars.put(name, Double.parseDouble(value));
              break;
            case VariablePO.TYPE_BOOL:
              vars.put(name, Boolean.parseBoolean(value));
              break;
//            case VariablePO.TYPE_DATE:
//              vars.put(name, Boolean.parseBoolean(value));
//              break;
            default:
              break;
          }
        } else if (variablePo.getType() == VariablePO.TYPE_CHAR) {
          vars.put(name, "");
        }
      }

      return vars;
    }

    return null;
  }

  public ProcessInstance startProcessInstanceByKey(long flowId, String flowNodeKey) {
    // 读取流程(新建流程)
    Flow flow = BpmFlowCache.getInstance().get(flowId);
    if (flow == null) {
      throw new BpmObjectNotFoundException("读取流程失败, model_id=" + flowId);
    }

    // 读取关联表
    ModelTable table = ModelTableCache.getInstance().get(flow.getTable().getId());
    if (table == null) {
      throw new BpmObjectNotFoundException("读取表失败, model_id=" + flow.getTable().getModelId() + ", table=" + flow.getTable().getName());
    }

    ProcessInstance processInstance = new ProcessInstance(dbEngine.nextId(), flow, table);
    FlowNode flowNode;

    if (StringUtils.isNotBlank(flowNodeKey)) {
      flowNode = flow.findNode(flowNodeKey);
    } else {
      flowNode = flow.getBeginNode();
    }

    if (flowNode == null) {
      throw new BpmObjectNotFoundException("未找到流程节点, model_id=" + flowId + ", nodeId=" + flowNodeKey);
    }

    processInstance.setCurNode(flowNode, false);

    return processInstance;
  }
}
