package cc.smtweb.system.bpm.engine.process.loader;

import cc.smtweb.framework.core.util.JsonUtil;
import cc.smtweb.system.bpm.core.exception.ModelLoaderError;
import cc.smtweb.system.bpm.engine.AbstractLoader;
import cc.smtweb.system.bpm.engine.process.loader.vo.*;
import cc.smtweb.system.bpm.engine.process.vo.*;
import org.apache.commons.lang3.StringUtils;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author xkliu
 * 流程定义读取器
 */
public class FlowLoader extends AbstractLoader<Flow> {
  private static final Map<String, Integer> PERM_MAP = new HashMap<>();
  private static final Map<String, Integer> ACTION_TYPE_MAP = new HashMap<>();
  private static final Map<String, Integer> USER_TYPE_MAP = new HashMap<>();
  private static final Map<String, Integer> USER_FILTER_MAP = new HashMap<>();

  static {
    PERM_MAP.put(NodeFieldVO.PERM_READ, FlowNodeField.PERM_READ);
    PERM_MAP.put(NodeFieldVO.PERM_WRITE, FlowNodeField.PERM_WRITE);
    PERM_MAP.put(NodeFieldVO.PERM_DISABLE, FlowNodeField.PERM_DISABLE);
    PERM_MAP.put(NodeFieldVO.PERM_HIDE, FlowNodeField.PERM_HIDE);

    ACTION_TYPE_MAP.put(ActionVO.TYPE_COMMIT, FlowAction.TYPE_COMMIT);
    ACTION_TYPE_MAP.put(ActionVO.TYPE_SAVE, FlowAction.TYPE_SAVE);
    ACTION_TYPE_MAP.put(ActionVO.TYPE_RETURN, FlowAction.TYPE_RETURN);
    ACTION_TYPE_MAP.put(ActionVO.TYPE_TRANSMIT, FlowAction.TYPE_TRANSMIT);
    ACTION_TYPE_MAP.put(ActionVO.TYPE_FINISH, FlowAction.TYPE_FINISH);
    ACTION_TYPE_MAP.put(ActionVO.TYPE_REJECT, FlowAction.TYPE_REJECT);

    USER_TYPE_MAP.put(NodeUsersVO.TYPE_ALL, FlowNodeUser.TYPE_ALL);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_ROLE, FlowNodeUser.TYPE_ROLE);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_USER_GROUP, FlowNodeUser.TYPE_USER_GROUP);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_USER, FlowNodeUser.TYPE_USER);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_DEPT, FlowNodeUser.TYPE_DEPT);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_FIELD, FlowNodeUser.TYPE_FIELD);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_EXP, FlowNodeUser.TYPE_EXP);
    USER_TYPE_MAP.put(NodeUsersVO.TYPE_API, FlowNodeUser.TYPE_API);

    USER_FILTER_MAP.put(NodeUsersVO.FILTER_CUR_DEPT, FlowNodeUser.FILTER_CUR_DEPT);
    USER_FILTER_MAP.put(NodeUsersVO.FILTER_SUB_DEPT, FlowNodeUser.FILTER_SUB_DEPT);
    USER_FILTER_MAP.put(NodeUsersVO.FILTER_UP_DEPT, FlowNodeUser.FILTER_UP_DEPT);
    USER_FILTER_MAP.put(NodeUsersVO.FILTER_CUR_USER, FlowNodeUser.FILTER_CUR_USER);
    USER_FILTER_MAP.put(NodeUsersVO.FILTER_CUR_MANAGER, FlowNodeUser.FILTER_CUR_MANAGER);
    USER_FILTER_MAP.put(NodeUsersVO.FILTER_UP_MANAGER, FlowNodeUser.FILTER_UP_MANAGER);
  }

  @Override
  public Flow load(String text) {
    clearErrors();

    if (StringUtils.isBlank(text)) {
      addError(ModelLoaderError.ERROR_FLOW, "流程为空");
      return null;
    }

    FlowVO flowVO = JsonUtil.parse(text, FlowVO.class);

    return convert(flowVO);
  }

  public Flow load(InputStream is) {
    clearErrors();

    if (is == null) {
      addError(ModelLoaderError.ERROR_FLOW, "流程为空");
      return null;
    }

    FlowVO flowVO = JsonUtil.parse(is, FlowVO.class);

    return convert(flowVO);
  }

  private Flow convert(FlowVO flowVO) {
    Flow result = new Flow();

    FlowPropsVO flowProps = flowVO.getProps();
    if (flowProps == null) {
      addError(ModelLoaderError.ERROR_FLOW, "流程缺少属性");
      return null;
    }

    if (flowProps.getDm() == null) {
      addError(ModelLoaderError.ERROR_FLOW, "流程缺少关联库");
      return null;
    }

    FlowTableVO flowTableVO =  flowProps.getTable();
    if (flowTableVO == null) {
      addError(ModelLoaderError.ERROR_FLOW, "流程缺少关联表");
      return null;
    }

    FlowTable flowTable = new FlowTable();
    flowTable.setModelId(flowProps.getDm().getModelId());
    flowTable.setId(flowTableVO.getId());
    flowTable.setName(flowTableVO.getName());
    result.setTable(flowTable);

    List<NodeVO> nodes = flowVO.getNodes();

    if (nodes == null) {
      addError(ModelLoaderError.ERROR_FLOW, "流程缺少活动节点");
      return null;
    }

    Map<String, FlowNode> nodeMap = new HashMap<>(nodes.size());
    FlowNode beginFlowNode = null;
    FlowNode endFlowNode = null;

    for (NodeVO nodeVO: nodes) {
      FlowNode flowNode = new FlowNode();
      flowNode.setId(nodeVO.getId());

      NodePropsVO nodePropsVO = nodeVO.getProps();

      if (nodePropsVO != null) {
        // 读取节点基本属性
        flowNode.setLabel(nodePropsVO.getLabel());
        flowNode.setName(nodePropsVO.getName());
        flowNode.setDisableNotify(nodePropsVO.getDisableNotify());

        switch (nodeVO.getType()) {
          case NodeVO.TYPE_BEGIN:
            flowNode.setType(FlowNode.TYPE_BEGIN);
            if (beginFlowNode == null) {
              beginFlowNode = flowNode;
            } else {
              addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "开始节点重复：" + beginFlowNode.getId());
            }
            break;
          case NodeVO.TYPE_END:
            flowNode.setType(FlowNode.TYPE_END);
            if (endFlowNode == null) {
              endFlowNode = flowNode;
            } else {
              addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "结束节点重复：" + endFlowNode.getId());
            }
            break;
          case NodeVO.TYPE_MANUAL:
            flowNode.setType(FlowNode.TYPE_MANUAL);
            break;
          case NodeVO.TYPE_MULTI_SIGN:
            flowNode.setType(FlowNode.TYPE_MULTI_SIGN);
            break;
          case NodeVO.TYPE_CONDITION:
            flowNode.setType(FlowNode.TYPE_CONDITION);
            break;
          case NodeVO.TYPE_SPLIT:
            flowNode.setType(FlowNode.TYPE_SPLIT);
            break;
          case NodeVO.TYPE_MERGE:
            flowNode.setType(FlowNode.TYPE_MERGE);
            break;
          case NodeVO.TYPE_SUB_FLOW:
            flowNode.setType(FlowNode.TYPE_SUB_FLOW);
            break;
          case NodeVO.TYPE_AUTO:
            flowNode.setType(FlowNode.TYPE_AUTO);
            break;
          default:
            addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "未支持的活动节点类型：" + nodeVO.getType());
            break;
        }

        // 读取节点字段权限
        List<NodeFieldVO> nodeFields = nodePropsVO.getFields();
        if (nodeFields != null) {
          List<FlowNodeField> fields = new ArrayList<>();

          for (NodeFieldVO nodeFieldVO: nodeFields) {
            FlowNodeField field = new FlowNodeField();
            field.setId(nodeFieldVO.getId());
            field.setLabel(nodeFieldVO.getLabel());
            field.setPerm(mapValue(PERM_MAP, nodeFieldVO.getPerm(), FlowNodeField.PERM_READ));
            field.setNotNull(nodeFieldVO.isNotNull());
            field.setDefaultValue(nodeFieldVO.getAssignValue());

            fields.add(field);
          }

          flowNode.setFields(fields);
        }

        // 读取动作和动作跳转
        List<ActionVO> actionVOs = nodeVO.getColumns();
        if (actionVOs != null) {
          List<FlowAction> flowActions = new ArrayList<>();
          for (ActionVO actionVO: actionVOs) {
            FlowAction flowAction = new FlowAction();

            flowAction.setId(actionVO.getId());
            flowAction.setType(mapValue(ACTION_TYPE_MAP, actionVO.getType()));

            ActionPropsVO actionPropsVO = actionVO.getProps();
            if (actionPropsVO != null) {
              flowAction.setLabel(actionPropsVO.getLabel());
            }

            flowActions.add(flowAction);

            List<ConnectionVO> connectionVOS = actionVO.getConnections();
            if (connectionVOS != null && connectionVOS.size() > 0) {
              ConnectionVO connectionVO = connectionVOS.get(0);

              FlowConnection flowConnection = new FlowConnection();
              flowConnection.setToNodeId(connectionVO.getTo());

              flowAction.setConnection(flowConnection);
            }
          }

          flowNode.setActions(flowActions);
        } else {
          if (flowNode.getType() != FlowNode.TYPE_END) {
            addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "活动节点缺少动作");
          }
        }

        // 执行对象
        List<NodeUsersVO> nodeUsersVos = nodePropsVO.getUsers();
        if (nodeUsersVos != null) {
          List<FlowNodeUser> flowNodeUsers = new ArrayList<>();
          for (NodeUsersVO nodeUsersVo: nodeUsersVos) {
            FlowNodeUser flowNodeUser = new FlowNodeUser();
            flowNodeUser.setType(mapValue(USER_TYPE_MAP, nodeUsersVo.getType()));
            flowNodeUser.setFilter(mapValue(USER_FILTER_MAP, nodeUsersVo.getFilter()));

            NodeUserTargetVO target = nodeUsersVo.getTarget();
            if (target != null && target.getUgId() != null) {
              flowNodeUser.setTarget(target.getUgId() + "");
            }

            flowNodeUsers.add(flowNodeUser);
          }

          flowNode.setUsers(flowNodeUsers);
        }
      } else {
        addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "活动节点缺少属性");
      }

      if (nodeMap.putIfAbsent(nodeVO.getId(), flowNode) != null) {
        addError(ModelLoaderError.ERROR_NODE, nodeVO.getId(), "活动节点ID重复：" + nodeVO.getId());
      }
    }

    // TODO 检查节点的连通性，死循环
    result.setNodeMap(nodeMap);
    result.setBeginNode(beginFlowNode);

    return result;
  }
}
