package cc.smtweb.system.bpm.engine.ui.codegen;

import cc.smtweb.framework.core.R;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.runtime.RuntimeConstants;
import org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader;

import java.io.*;

public abstract class AbstractVelocityEngine {
  private static final int DIGEST_SIZE = 9+64+9+17;
  private static final int MIN_FILE_SIZE = DIGEST_SIZE + 4;
  private static final int MAX_FILE_SIZE = 1024 * 1024 *10;

  private VelocityEngine ve = new VelocityEngine();
  protected VelocityContext ctx;

  public AbstractVelocityEngine() {
    ve.setProperty(RuntimeConstants.RESOURCE_LOADER, "classpath");
    ve.setProperty("classpath.resource.loader.class", ClasspathResourceLoader.class.getName());
    ve.init();
  }

  /** 生成文件，通过签名检查就替换文件 */
  public R buildUseSign(long id, String templateFile, File destFile) throws IOException {
    // 载入（获取）模板对象
    Template t = ve.getTemplate(templateFile);
//    VelocityContext ctx = getVelocityContext(uiPage, uiPageRef);
    File tmpFile = new File(destFile.getAbsoluteFile() + ".tmp");

    Sha256OutputStream shaOutputStream = new Sha256OutputStream(new FileOutputStream(tmpFile), true);

    try(Writer writer = new OutputStreamWriter(shaOutputStream)) {
      t.merge(ctx, writer);

      writer.flush();
      String sha256 = shaOutputStream.sha256Hex();
      int digestSize = shaOutputStream.digestSize();

      // 写入11+64+9+17字节签名 <!-- [sha256] [Hex(Size)] [Hex(id)] -->
      if (destFile.getName().endsWith(".vue")) {
        writer.write(String.format("\n<!-- %s,%08x,%016x -->\n", sha256, digestSize, id));
      } else { // ts / java
        writer.write(String.format("\n/* ![%s,%08x,%016x] */\n", sha256, digestSize, id));
      }
    }

    // 签名对比是否手动修改
    if (destFile.exists()) {
      R r = checkOldFile(id, destFile);
      if (r != null) {
        return r;
      }

      destFile.delete();
    }

    // 更名正式文件
    if (!tmpFile.renameTo(destFile)) {
      return R.error("写文件失败，请用tmp文件手动替换！");
    }

    return null;
  }

  /** 生成文件，文件不存在就替换文件 */
  public R build(String templateFile, File destFile) throws IOException {
    if (!destFile.exists()) {
      // 载入（获取）模板对象
      Template t = ve.getTemplate(templateFile);
//    VelocityContext ctx = getVelocityContext(uiPage, uiPageRef);

      try(Writer writer = new OutputStreamWriter(new FileOutputStream(destFile))) {
        t.merge(ctx, writer);
      }
    }

    return null;
  }

  protected R checkOldFile(Long modelId, File vmFile) throws IOException {
    // TODO 判断签名是否可以覆盖
    long vmFileLength = vmFile.length();
    if (vmFileLength <= MIN_FILE_SIZE) {
      return R.error("文件太小：" + vmFileLength);
    }

    if (vmFileLength > MAX_FILE_SIZE) {
      return R.error("文件太大：" + vmFileLength);
    }

    try (RandomAccessFile raFile = new RandomAccessFile(vmFile, "r")) {
      raFile.seek(vmFile.length() - MIN_FILE_SIZE);
      byte[] digest = new byte[MIN_FILE_SIZE];
      if (raFile.read(digest) != MIN_FILE_SIZE) {
        return R.error("读取签名数据失败");
      }

      // 查找签名行
      String[] tokens = null;
      int digestEnd = 0;
      for (int i = digest.length - 1; i >= 0; i--) {
        byte b = digest[i];
        if (b >= 0 && b <= 20) {
          // 换行符
          if (digestEnd > 0) {
            if (digestEnd - i == DIGEST_SIZE) {
              String line = new String(digest, i + 1, digestEnd - i);
              tokens = line.substring(5, line.length() - 4).split(",");
              vmFileLength -= digest.length - i;
            }

            break;
          }
        } else {
          // 有效字符
          if (digestEnd == 0) {
            digestEnd = i;
          }
        }
      }

      if (tokens == null) {
        return R.error("未找到签名数据");
      }

//        String[] tokens = line.substring(6, line.length() - 5).split(",");
      if (tokens.length < 3) {
        return R.error("签名格式出错");
      }
      // sha256,size,id
      long id = Long.parseLong(tokens[2].trim(), 16);
      if (id != modelId) {
        return R.error("文件ID值不一致：" + id);
      }

      int size = (int) (vmFileLength); //  - DIGEST_SIZE

      raFile.seek(0);
      byte[] fileData = new byte[size];
      if (raFile.read(fileData) != size) {
        return R.error("读取文件失败");
      }

      Sha256Helper sha256Helper = new Sha256Helper(true);
      sha256Helper.write(fileData);
      String sha256 = sha256Helper.sha256Hex();

      if (!sha256.equals(tokens[0].trim())) {
        return R.error("文件签名比对失败，文件内容请参考tmp文件手动调整！");
      }

      int oldSize = Integer.parseInt(tokens[1].trim(), 16);
      if (oldSize != sha256Helper.digestSize()) {
        return R.error("文件尺寸校验失败，文件内容请参考tmp文件手动调整！");
      }
    }

    return null;
  }

  private String[] findDigestLine(byte[] digest) {
    int stop = 0;
    for (int i = digest.length - 1; i >= 0; i--) {
      byte b = digest[i];
      if (stop > 0 && b >= 0 && b <= 20) {
        if (stop - i == DIGEST_SIZE) {
          String line = new String(digest, i, stop - i);
//        String[] tokens = line.substring(6, line.length() - 5).split(",");
          return line.substring(5, line.length() - 4).split(",");
        } else {
          return null;
        }
      } else if (b < 0 || b > 20) {
        stop = i;
      }
    }

    return null;
  }

}
