package cc.smtweb.system.bpm.engine.ui.loader.yaml;

import cc.smtweb.system.bpm.core.exception.BpmException;
import cc.smtweb.system.bpm.core.ui.entity.dataset.*;
import cc.smtweb.system.bpm.engine.ui.entity.consts.MappingTypeEnum;
import cc.smtweb.system.bpm.engine.ui.entity.consts.UiEnum;
import cc.smtweb.system.bpm.engine.ui.entity.yaml.*;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.List;

public class YamlDatasetDecoder {
  private YamlDatasetDecoder() {}

  public static BpmDataset decode(YamlDataset yamlDataset) {
    BpmDataset bpmDataset = new BpmDataset();
    bpmDataset.setName(yamlDataset.getName());
    bpmDataset.setDatabase(yamlDataset.getDatabase());
    bpmDataset.setTable(yamlDataset.getTable());
    bpmDataset.setDataModel(decodeDataModel(yamlDataset.getDataModel()));
    bpmDataset.setReadonly(yamlDataset.isReadonly());
    bpmDataset.setCustomApi(StringUtils.trimToNull(yamlDataset.getCustomApi()));
    bpmDataset.setInitData(decodeBoolean(yamlDataset.getInitData()));
    bpmDataset.setSql(decodeSql(yamlDataset.getSql()));
    decodeMasterMapping(bpmDataset.getMasterMapping(), yamlDataset.getMasterMapping());

    yamlDataset.getFields().forEach((field) -> {
      BpmField bpmField = new BpmField();

      bpmField.setType(field.getType());
      bpmField.setName(field.getName());
      bpmField.setLabel(field.getLabel());
      bpmField.setDisplayFormat(field.getDisplayFormat());
      bpmField.setFieldName(field.getFieldName());
      bpmField.setSummary(field.getSummary());
      bpmField.setLookup(decodeLookup(field.getLookup()));
      bpmField.setRefType(decodeRefType(field.getRefType()));
      bpmField.setRequired(decodeBoolean(field.getRequired()));
      bpmField.setReadonly(decodeBoolean(field.getReadonly()));
      bpmField.setValidPatten(StringUtils.trimToNull(field.getValidPatten()));
      bpmField.setValidMessage(StringUtils.trimToNull(field.getValidMessage()));
      bpmField.setFilter(field.getFilter());
      bpmField.setFilterType(field.getFilterType());

      bpmDataset.addField(bpmField);
    });

    bpmDataset.setFieldPrefix(getFieldPrefix(bpmDataset.readIdFieldName()));

    return bpmDataset;
  }

  private static BpmDatasetSql decodeSql(YamlDatasetSql sql) {
    if (sql != null && StringUtils.isNotBlank(sql.getSelect())) {
      BpmDatasetSql result = new BpmDatasetSql();
      result.setSelect(sql.getSelect());

      if (StringUtils.isNotBlank(sql.getInsert())) {
        result.setInsert(sql.getInsert());
      }

      if (StringUtils.isNotBlank(sql.getUpdate())) {
        result.setUpdate(sql.getUpdate());
      }

      return result;
    }

    return null;
  }

  private static boolean decodeBoolean(Boolean value) {
    return Boolean.TRUE.equals(value);
  }

  private static int decodeRefType(String refType) {
    Integer value = UiEnum.FieldRefType.INSTANCE.getKey(refType);

    if (value != null) {
      return value;
    }

    return 0;
  }

  private static void decodeMasterMapping(BpmMasterMapping bpmMasterMapping, YamlMasterMapping masterMapping) {
    if (masterMapping != null && masterMapping.getMappingFields() != null && !MappingTypeEnum.none.equals(masterMapping.getMappingType())) {
      bpmMasterMapping.setMappingType(MappingTypeEnum.toEnum(masterMapping.getMappingType()));
      bpmMasterMapping.setMaster(masterMapping.getMasterTable());
      bpmMasterMapping.setMiddle(masterMapping.getMiddleTable());

      List<BpmMappingField> mappingFields = new ArrayList<>(masterMapping.getMappingFields().size());
      for (YamlMappingField mappingField: masterMapping.getMappingFields()) {
        BpmMappingField bpmMappingField = new BpmMappingField();
        bpmMappingField.setMaster(mappingField.getMasterField());
        bpmMappingField.setSlave(mappingField.getSlaveField());
        bpmMappingField.setMiddle(mappingField.getMiddleField());

        mappingFields.add(bpmMappingField);
      }

      bpmMasterMapping.setMappingFields(mappingFields);
    }
  }

  private static int decodeDataModel(String dataModel) {
    switch (dataModel) {
      case "card":
        return BpmDataset.DATA_MODEL_CARD;
      case "list":
        return BpmDataset.DATA_MODEL_LIST;
      case "tree":
        return BpmDataset.DATA_MODEL_TREE;
      default:
        throw new BpmException(String.format("不支持的refType类型:%s", dataModel));
    }
  }

  protected static AbstractBpmFieldLookup decodeLookup(YamlFieldLookup lookup) {
    if (lookup != null) {
      int lookupType = UiEnum.FieldLookupType.INSTANCE.getKey(lookup.getType());
      if (lookupType == UiEnum.FieldLookupType.DICT) {
        String dictType = lookup.getFilters().get(0).getKeyword();
        return new BpmFieldDict(lookup.getDb(), lookup.getTable(), lookup.isNumberValue(), dictType);
      } else if (lookupType == UiEnum.FieldLookupType.TABLE) {
        BpmFieldLookup bpmFieldLookup = new BpmFieldLookup(lookup.getDb(), lookup.getTable(), lookup.isNumberValue(), lookup.getMappingType());
        bpmFieldLookup.setKeyField(lookup.getKeyField());
        bpmFieldLookup.setLabelField(lookup.getLabelField());
        bpmFieldLookup.setOrderField(lookup.getOrderField());
        bpmFieldLookup.setParentField(lookup.getParentField());

        if (bpmFieldLookup.isManyToMany()) {
          YamlFieldLookupMiddle middle  = lookup.getMiddle();
          bpmFieldLookup.setMiddleTable(middle.getTable());
          bpmFieldLookup.setMiddleIdField(middle.getIdField());
          bpmFieldLookup.setMiddleKeyField(middle.getKeyField());
          bpmFieldLookup.setMiddleValueField(middle.getValueField());
        }

        if (lookup.getFilters() != null) {
          for (YamlFieldFilter yamlFieldFilter : lookup.getFilters()) {
            bpmFieldLookup.addFilter(yamlFieldFilter.getField(), yamlFieldFilter.getOp(), yamlFieldFilter.getKeyword(), yamlFieldFilter.isCascade());
          }
        }

        return bpmFieldLookup;
      }
    }

    return null;
  }

  public static String getFieldPrefix(String idField) {
    if (idField != null) {
      String[] ary = idField.split("_");
      if (ary.length > 0) {
        return ary[0];
      }
    }

    return null;
  }
}
