package cc.smtweb.system.bpm.engine.ui.loader.yaml;

import cc.smtweb.system.bpm.core.exception.BpmException;
import cc.smtweb.system.bpm.core.ui.entity.dataset.*;
import cc.smtweb.system.bpm.engine.ui.entity.consts.MappingTypeEnum;
import cc.smtweb.system.bpm.engine.ui.entity.consts.UiEnum;
import cc.smtweb.system.bpm.engine.ui.entity.yaml.*;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.List;

public class YamlDatasetEncoder {
  private BpmDataset dataset;

  public YamlDatasetEncoder(BpmDataset dataset) {
    this.dataset = dataset;
  }

  public YamlDataset encode() {
    YamlDataset yamlDataset = new YamlDataset();
    yamlDataset.setName(dataset.getName());
    yamlDataset.setDatabase(dataset.getDatabase());
    yamlDataset.setTable(dataset.getTable());
    yamlDataset.setDataModel(encodeDataModel(dataset.getDataModel()));
    yamlDataset.setReadonly(dataset.isReadonly());
    yamlDataset.setCustomApi(StringUtils.trimToNull(dataset.getCustomApi()));
    yamlDataset.setInitData(encodeBoolean(dataset.isInitData()));
    yamlDataset.setMasterMapping(encodeMasterMapping(dataset.getMasterMapping()));
    yamlDataset.setSql(encodeSql(dataset.getSql()));

    List<YamlField> yamlFields = new ArrayList<>();
    dataset.getFields().forEach((name, field) -> {
      YamlField yamlField = new YamlField();

      yamlField.setType(field.getType());
      yamlField.setName(field.getName());
      yamlField.setLabel(field.getLabel());
      yamlField.setDisplayFormat(field.getDisplayFormat());
      yamlField.setFieldName(field.getFieldName());
      yamlField.setSummary(field.getSummary());
      yamlField.setLookup(encodeLookup(field.getLookup()));
      yamlField.setRefType(encodeRefType(field.getRefType()));
      yamlField.setRequired(encodeBoolean(field.isRequired()));
      yamlField.setReadonly(encodeBoolean(field.isReadonly()));
      yamlField.setValidPatten(StringUtils.trimToNull(field.getValidPatten()));
      yamlField.setValidMessage(StringUtils.trimToNull(field.getValidMessage()));
      yamlField.setFilter(field.getFilter());
      yamlField.setFilterType(field.getFilterType());

      yamlFields.add(yamlField);
    });

    yamlDataset.setFields(yamlFields);

    return yamlDataset;
  }

  private YamlDatasetSql encodeSql(BpmDatasetSql sql) {
    if (sql != null && StringUtils.isNotBlank(sql.getSelect())) {
      YamlDatasetSql result = new YamlDatasetSql();
      result.setSelect(sql.getSelect());

      if (StringUtils.isNotBlank(sql.getInsert())) {
        result.setInsert(sql.getInsert());
      }

      if (StringUtils.isNotBlank(sql.getUpdate())) {
        result.setUpdate(sql.getUpdate());
      }

      return result;
    }

    return null;
  }

  private Boolean encodeBoolean(boolean value) {
    return value ? Boolean.TRUE : null;
  }

  private YamlMasterMapping encodeMasterMapping(BpmMasterMapping masterMapping) {
    if (masterMapping != null && masterMapping.getMappingFields() != null && !MappingTypeEnum.none.equals(masterMapping.getMappingType())) {
      YamlMasterMapping yamlMasterMapping = new YamlMasterMapping();
      yamlMasterMapping.setMappingType(masterMapping.getMappingType().name());
      yamlMasterMapping.setMasterTable(masterMapping.getMaster());
      yamlMasterMapping.setMiddleTable(masterMapping.getMiddle());

      List<YamlMappingField> mappingFields = new ArrayList<>(masterMapping.getMappingFields().size());
      for (BpmMappingField mappingField: masterMapping.getMappingFields()) {
        YamlMappingField yamlMappingField = new YamlMappingField();
        yamlMappingField.setMasterField(mappingField.getMaster());
        yamlMappingField.setSlaveField(mappingField.getSlave());
        yamlMappingField.setMiddleField(mappingField.getMiddle());

        mappingFields.add(yamlMappingField);
      }

      yamlMasterMapping.setMappingFields(mappingFields);

      return yamlMasterMapping;
    }

    return null;
  }

  private YamlFieldLookup encodeLookup(AbstractBpmFieldLookup lookup) {
    YamlFieldLookup yamlFieldLookup = null;
    if (lookup != null) {
      String lookupType = UiEnum.FieldLookupType.INSTANCE.getValue(lookup.type());
      yamlFieldLookup = new YamlFieldLookup();
      yamlFieldLookup.setType(lookupType);
      yamlFieldLookup.setNumberValue(lookup.isNumber());
      yamlFieldLookup.setMappingType(lookup.getMappingType());
      yamlFieldLookup.setDb(lookup.getDb());
      yamlFieldLookup.setTable(lookup.getTable());

      if (lookup.type() == UiEnum.FieldLookupType.DICT) {
        BpmFieldDict bpmFieldDict = (BpmFieldDict)lookup;
        yamlFieldLookup.setKeyField("dict_code");
        yamlFieldLookup.setLabelField("dict_label");
        yamlFieldLookup.setOrderField("dict_order");

        List<YamlFieldFilter> yamlFieldFilters = new ArrayList<>(1);
        YamlFieldFilter yamlFieldFilter = new YamlFieldFilter();
        yamlFieldFilter.setField("dt_name");
        yamlFieldFilter.setKeyword(bpmFieldDict.getDictType());
        yamlFieldFilters.add(yamlFieldFilter);

        yamlFieldLookup.setFilters(yamlFieldFilters);
      } else if (lookup.type() == UiEnum.FieldLookupType.TABLE) {
        BpmFieldLookup bpmFieldLookup = (BpmFieldLookup)lookup;
        yamlFieldLookup.setKeyField(bpmFieldLookup.getKeyField());
        yamlFieldLookup.setLabelField(bpmFieldLookup.getLabelField());
        yamlFieldLookup.setOrderField(bpmFieldLookup.getOrderField());
        yamlFieldLookup.setParentField(bpmFieldLookup.getParentField());

        if (bpmFieldLookup.isManyToMany()) {
          YamlFieldLookupMiddle middle  = new YamlFieldLookupMiddle();
          middle.setTable(bpmFieldLookup.getMiddleTable());
          middle.setIdField(bpmFieldLookup.getMiddleIdField());
          middle.setKeyField(bpmFieldLookup.getMiddleKeyField());
          middle.setValueField(bpmFieldLookup.getMiddleValueField());

          yamlFieldLookup.setMiddle(middle);
        }

        if (bpmFieldLookup.getFilters() != null) {
          List<YamlFieldFilter> yamlFieldFilters = new ArrayList<>();
          for (BpmFieldFilter filter : bpmFieldLookup.getFilters()) {
            YamlFieldFilter yamlFieldFilter = new YamlFieldFilter();
            yamlFieldFilter.setField(filter.getField());
            yamlFieldFilter.setKeyword(filter.getKeyword());
            yamlFieldFilter.setOp(filter.getOp());
            yamlFieldFilter.setCascade(filter.isCascade());
            yamlFieldFilters.add(yamlFieldFilter);
          }
          yamlFieldLookup.setFilters(yamlFieldFilters);
        }
      }
    }

    return yamlFieldLookup;
  }

  private String encodeRefType(int refType) {
    if (refType == 0) {
      return null;
    }

    String result = UiEnum.FieldRefType.INSTANCE.getValue(refType);

    if (result != null) {
      return result;
    }

    throw new BpmException(String.format("%s不支持的refType类型:%d", dataset.getName(), refType));
  }

  private String encodeDataModel(int dataModel) {
    switch (dataModel) {
      case BpmDataset.DATA_MODEL_CARD:
        return "card";
      case BpmDataset.DATA_MODEL_LIST:
        return "list";
      case BpmDataset.DATA_MODEL_TREE:
        return "tree";
      case 0:
        throw new BpmException(String.format("未设置%s.dataModel类型", dataset.getName()));
      default:
        throw new BpmException(String.format("不支持的%s.dataModel类型:%d", dataset.getName(), dataModel));
    }
  }
}
