package cc.smtweb.system.bpm.spring;

import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.system.bpm.core.exception.BpmIlegalArgumentException;
import cc.smtweb.system.bpm.core.exception.BpmObjectNotFoundException;
import cc.smtweb.system.bpm.engine.process.impl.dao.TaskDao;
import cc.smtweb.system.bpm.engine.process.impl.entity.CommitTaskUserVO;
import cc.smtweb.system.bpm.engine.process.runtime.*;
import cc.smtweb.system.bpm.engine.process.task.Task;
import cc.smtweb.system.bpm.engine.process.vo.Flow;
import cc.smtweb.system.bpm.engine.process.vo.FlowAction;
import cc.smtweb.system.bpm.engine.process.vo.FlowConnection;
import cc.smtweb.system.bpm.engine.process.vo.FlowNode;
import cc.smtweb.system.bpm.engine.ui.IFormDataStore;
import cc.smtweb.system.bpm.engine.ui.impl.FormDataStoreImpl;
import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;

import java.sql.Timestamp;
import java.util.List;

@Service
public class ProcessEngineImpl implements IProcessEngine {
  private DbEngine dbEngine;
  private TaskDao taskDao;

  private IFormDataStore formDataStore;

  public ProcessEngineImpl(DbEngine dbEngine) {
    taskDao = new TaskDao(dbEngine);
    this.dbEngine = dbEngine;
    this.formDataStore = new FormDataStoreImpl(dbEngine);
  }

  @Override
  public Task newTask(long flowId, String nodeKey) {
    ProcessInstance processInstance = taskDao.startProcessInstanceByKey(flowId, nodeKey);

    FlowNode flowNode;

    if (nodeKey == null) {
      // 找到第一个执行节点，读取字段权限
      flowNode = processInstance.nextNode(FlowAction.TYPE_COMMIT);
    } else {
      flowNode = processInstance.getFlow().findNode(nodeKey);
    }

    if (flowNode == null) {
      throw new BpmObjectNotFoundException("not find node by key: " + nodeKey);
    }

    processInstance.setCurNode(flowNode, nodeKey == null);
    return new Task(dbEngine.nextId(), processInstance);
  }

  @Override
  public Task saveTask(Task task) {
    Timestamp now = new Timestamp(System.currentTimeMillis());

    ProcessInstance processInstance = task.getProcessInstance();
    taskDao.saveProcess(processInstance, now);
    taskDao.saveTask(task, now);

    return task;
  }

  private CommitTaskUserVO findTaskUser(List<CommitTaskUserVO> users, String nodeKey) {
    for (CommitTaskUserVO user : users) {
      // 判断flowNode的合法性和必填性
      if (nodeKey.equals(user.getNodeKey())) {
        return user;
      }
    }

    return null;
  }

  @Override
  public void commitTask(Task task, FlowAction flowAction, List<CommitTaskUserVO> users) {
    switch (flowAction.getType()) {
      case FlowAction.TYPE_COMMIT:
        FlowNode flowNode = task.getProcessInstance().getFlow().checkAndGetNode(flowAction.checkAndGetConnection().getToNodeId());
        RouteInfo nextProcessInfo = this.routeCommit(task, flowNode);
        commitRoute(task, nextProcessInfo, users);
        break;
      case FlowAction.TYPE_TRANSMIT:
        // 转发
        CommitTaskUserVO commitTaskUser = users.get(0);
        task.setAssigneeUserId(commitTaskUser.getUserId());
        saveTask(task);
        break;
      case FlowAction.TYPE_RETURN:
        // 1. 终止所有同级子流程
        // 2. 回到来源节点（可能是多个）
        nextProcessInfo = this.routeReturn(task);
        commitRoute(task, nextProcessInfo, users);
        break;
      default:
        throw new BpmIlegalArgumentException("error action type: " + flowAction.getType());
    }
  }

  public void commitRoute(Task task, RouteInfo nextProcessInfo, List<CommitTaskUserVO> users) {
    ProcessInstance processInstance = task.getProcessInstance();
    Timestamp now = new Timestamp(System.currentTimeMillis());

    dbEngine.doTrans(() -> {
      for (Route route: nextProcessInfo.getRoutes()) {
        FlowNode flowNode = route.getNode();
        if (flowNode != null) {
          CommitTaskUserVO user = findTaskUser(users, flowNode.getId());

          // 判断flowNode的合法性和必填性
          if (user == null) {
            throw new BpmObjectNotFoundException("not find task user by node key: " + flowNode.getId());
          }

          Long userId = user.getUserId();
          Task nextTask;

          switch (flowNode.getType()) {
            case FlowNode.TYPE_END:
            case FlowNode.TYPE_MERGE:
              break;
            case FlowNode.TYPE_MANUAL:
              if (userId == null) {
                throw new BpmIlegalArgumentException("not set assignee user");
              }
            default:
              ProcessInstance subProcess = route.getProcess();
              nextTask = new Task(dbEngine.nextId(), subProcess);

              nextTask.setAssigneeUserId(userId);
              this.saveTask(nextTask);
              break;
          }
        } else {
          taskDao.deleteProcess(route.getProcess());
        }
      }

      taskDao.deleteTask(task);

      return true;
    });
  }

  @Override
  public void deleteTask(Task task) {
    taskDao.deleteTask(task);
  }

  @Override
  public Task loadTask(long taskId) {
    return taskDao.loadTask(taskId);
  }

  @Override
  public RouteInfo routeCommit(Task task, FlowNode toFlowNode) {
    RouteInfo result = new RouteInfo();
    fillNextNodes(result, task.getProcessInstance(), toFlowNode);
    return result;
  }

  @Override
  public IFormDataStore getFormDataStore() {
    return formDataStore;
  }

  private void fillNextNodes(RouteInfo result, ProcessInstance processInstance, FlowNode toFlowNode) {
    Flow flow = processInstance.getFlow();

    switch (toFlowNode.getType()) {
      case FlowNode.TYPE_END:
        result.add(processInstance, null);
      case FlowNode.TYPE_AUTO:
      case FlowNode.TYPE_MANUAL:
        result.add(processInstance, toFlowNode);
        break;
      case FlowNode.TYPE_SPLIT: {
        List<FlowAction> flowActions = toFlowNode.getActions();
        if (flowActions != null) {
          for (FlowAction splitFlowAction : flowActions) {
            FlowConnection splitFlowConnection = splitFlowAction.checkAndGetConnection();
            FlowNode splitFlowNode = flow.checkAndGetNode(splitFlowConnection.getToNodeId());

            // create subProcess
            ProcessInstance subProcess = processInstance.createSubProcess(dbEngine.nextId(), splitFlowNode);
            subProcess.setConcurrent(true);
            fillNextNodes(result, subProcess, splitFlowNode);
          }
        }
      }
        break;
      case FlowNode.TYPE_MERGE:
        // 结束并行流程
        result.add(processInstance, null);

        ProcessInstance parentProcess = taskDao.loadProcess(processInstance.getParentId());

        if (parentProcess == null) {
          throw new BpmObjectNotFoundException("not find process by id " + processInstance.getParentId());
        }

        ProcessRouteItem processRouteItem = parentProcess.getProcessRoute().findRouteByProcess(processInstance.getId());

        if (processRouteItem == null) {
          throw new BpmObjectNotFoundException("not find route by process id " + processInstance.getId());
        }

        // TODO 调整为通过路由列表来判断是否最后一个执行人
        if (dbEngine.queryInt("select count(*) from sw_bpm.bpm_task where task_process_id in (" + StringUtils.join(processRouteItem.getSub(), ",") + ")") == 1) {
          FlowAction nextFlowAction = toFlowNode.findAction(FlowAction.TYPE_COMMIT);

          if (nextFlowAction == null) {
            throw new BpmObjectNotFoundException("未找到下一步Action, node_id=" + toFlowNode.getId());
          }

          // 回到上级流程继续执行
          FlowConnection nextFlowConnection = nextFlowAction.checkAndGetConnection();
          FlowNode nextFlowNode = flow.checkAndGetNode(nextFlowConnection.getToNodeId());

          fillNextNodes(result, parentProcess, nextFlowNode);
        }
        break;
      case FlowNode.TYPE_CONDITION: {
        List<FlowAction> flowActions = toFlowNode.getActions();
        if (flowActions != null) {
          for (FlowAction action : flowActions) {
            if (checkCondition(action, processInstance)) {
              FlowConnection splitFlowConnection = action.checkAndGetConnection();
              FlowNode splitFlowNode = flow.checkAndGetNode(splitFlowConnection.getToNodeId());
              fillNextNodes(result, processInstance, splitFlowNode);
              break;
            }
          }
        }
      }
        break;
      default:
        break;
    }
  }

  @Override
  public RouteInfo routeReturn(Task task) {
    RouteInfo result = new RouteInfo();

    task.getProcessInstance().getProcessRoute().returnRoute();
    fillReturnNodes(result, task.getProcessInstance(), true);

    return result;
  }

  private void fillReturnNodes(RouteInfo result, ProcessInstance processInstance, boolean resume) {
    ProcessRouteItem routeItem = processInstance.getProcessRoute().lastRoute();
    if (routeItem != null) {
      if (routeItem.getNode() != null) {
        // 恢复一个节点
        result.add(processInstance, processInstance.getFlow().checkAndGetNode(routeItem.getNode()));
      } else {
        List<Long> ids = routeItem.getSub();
        if (resume) {
          // 恢复并行流程
          for (int i = 0; i < ids.size(); i++) {
            // 到历史记录中读取子流程
            ProcessInstance subProcess = taskDao.loadProcessByLog(ids.get(i));
            if (subProcess == null) {
              throw new BpmObjectNotFoundException("not find process log by id " + ids.get(i));
            }

            subProcess.resume(dbEngine.nextId());
            ids.set(i, subProcess.getId());
            fillReturnNodes(result, subProcess, true);
          }
        } else {
          // 停止同级并行流程
          List<ProcessInstance> processList = taskDao.loadProcessList(ids);
          if (processList != null) {
            for (ProcessInstance pi: processList) {
              result.add(pi, null);
            }
          }
        }
      }
    } else if (processInstance.getParentId() != null) {
      // 读取上级流程
      ProcessInstance parentProcess = taskDao.loadProcess(processInstance.getParentId());
      fillReturnNodes(result, parentProcess, false);
    }
  }

  /** SpEL 表达式计算 */
  private boolean checkCondition(FlowAction action, ProcessInstance processInstance) {
    String expression = action.getExpression();
    if (StringUtils.isBlank(expression)) {
      return true;
    }

    ConditionExpression conditionExpression = processInstance.getConditionExpression();
    if (conditionExpression == null) {
      ProcessContext rootObject = createProcessContext(processInstance);

      conditionExpression = new ConditionExpression(rootObject);
      processInstance.setConditionExpression(conditionExpression);
    }

    Boolean value = conditionExpression.execute(expression, Boolean.class);
    return value != null ? value: false;
  }

  private ProcessContext createProcessContext(ProcessInstance processInstance) {
    ProcessContext context = new ProcessContext();

    context.setVars(taskDao.loadVars(processInstance));

    // 读取表单数据
    context.setData(this.formDataStore.load(processInstance.getId(), processInstance.getRootId()));

    return context;
  }
}
