package cc.smtweb.system.bpm.spring.action;

import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.mvc.controller.access.IMethodAccess;
import cc.smtweb.system.bpm.core.annotation.BpmEvent;
import cc.smtweb.system.bpm.core.annotation.BpmEventType;
import cc.smtweb.system.bpm.core.ui.BpmPageContext;
import cc.smtweb.system.bpm.core.ui.service.DatasetApiInvoker;
import cc.smtweb.system.bpm.core.ui.service.card.AbstractDatasetCardService;
import cc.smtweb.system.bpm.core.ui.service.list.AbstractDatasetListService;
import cc.smtweb.system.bpm.core.ui.service.tree.AbstractDatasetTreeService;
import cc.smtweb.system.bpm.spring.action.service.BpmActionCardService;
import cc.smtweb.system.bpm.spring.action.service.BpmActionListService;
import cc.smtweb.system.bpm.spring.action.service.BpmActionTreeService;
import cc.smtweb.system.bpm.spring.dao.DatasetCardDao;
import cc.smtweb.system.bpm.spring.dao.DatasetConfigDao;
import cc.smtweb.system.bpm.spring.dao.DatasetListDao;
import cc.smtweb.system.bpm.spring.dao.DatasetTreeDao;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.HashMap;
import java.util.Map;

/**
 * BpmEvent事件处理Dao
 * @author xkliu
 */
@Slf4j
@Service
public class BpmEventDao {
  private DbEngine dbEngine;
  private DatasetConfigDao datasetConfigDao;
  private DatasetCardDao datasetCardDao;
  private DatasetListDao datasetListDao;
  private DatasetTreeDao datasetTreeDao;

  private Map<String, BpmActionCardService> cardServiceMap = new HashMap<>();
  private Map<String, BpmActionListService> listServiceMap = new HashMap<>();
  private Map<String, BpmActionTreeService> treeServiceMap = new HashMap<>();

  private BpmActionCardService defaultActionCardService;
  private BpmActionListService defaultActionListService;
  private BpmActionTreeService defaultActionTreeService;

  public BpmEventDao(DbEngine dbEngine, DatasetConfigDao datasetConfigDao,
                     DatasetCardDao datasetCardDao, DatasetListDao datasetListDao, DatasetTreeDao datasetTreeDao) {
    this.dbEngine = dbEngine;
    this.datasetConfigDao = datasetConfigDao;
    this.datasetCardDao = datasetCardDao;
    this.datasetListDao = datasetListDao;
    this.datasetTreeDao = datasetTreeDao;

    defaultActionCardService = new BpmActionCardService(dbEngine, datasetConfigDao, datasetCardDao, null);
    defaultActionListService = new BpmActionListService(dbEngine, datasetConfigDao, datasetListDao, null);
    defaultActionTreeService = new BpmActionTreeService(dbEngine, datasetConfigDao, datasetTreeDao, null);
  }

  // 系统启动时添加扫描的BpmEvent
  // 检测是否重复设置和相同动作时类型匹配
  void addAction(BpmEvent bpmEvent, IMethodAccess methodAccess) {
    switch (bpmEvent.event()) {
      case BpmEventType.CARD_BEFORE_LOAD:
        buildCardService(bpmEvent).setBeforeLoad(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_LOAD: {
        buildCardService(bpmEvent).setLoad(new DatasetApiInvoker(methodAccess));
      }
        break;
      case BpmEventType.CARD_AFTER_LOAD:
        buildCardService(bpmEvent).setAfterLoad(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_BEFORE_SAVE:
        buildCardService(bpmEvent).setBeforeSave(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_SAVE:
        buildCardService(bpmEvent).setSave(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_AFTER_SAVE:
        buildCardService(bpmEvent).setAfterSave(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_BEFORE_REMOVE:
        buildCardService(bpmEvent).setBeforeRemove(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_REMOVE:
        buildCardService(bpmEvent).setRemove(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.CARD_AFTER_REMOVE:
        buildCardService(bpmEvent).setAfterRemove(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_BEFORE_LIST:
        buildListService(bpmEvent).setBeforeList(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_LIST:
        buildListService(bpmEvent).setList(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_AFTER_LIST:
        buildListService(bpmEvent).setAfterList(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_BEFORE_COUNT:
        buildListService(bpmEvent).setBeforeCount(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_AFTER_COUNT:
        buildListService(bpmEvent).setAfterCount(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.LIST_COUNT:
        buildListService(bpmEvent).setCount(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_BEFORE_TREE:
        buildTreeService(bpmEvent).setBeforeTreeAll(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_AFTER_TREE:
        buildTreeService(bpmEvent).setAfterTreeAll(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_TREE:
        buildTreeService(bpmEvent).setTreeAll(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_BEFORE_MOVE:
        buildTreeService(bpmEvent).setBeforeMove(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_AFTER_MOVE:
        buildTreeService(bpmEvent).setAfterMove(new DatasetApiInvoker(methodAccess));
        break;
      case BpmEventType.TREE_MOVE:
        buildTreeService(bpmEvent).setMove(new DatasetApiInvoker(methodAccess));
        break;
      default:
        log.error("unknown type :" + bpmEvent.event());
        break;
    }
  }

  private BpmActionCardService buildCardService(BpmEvent bpmEvent) {
    BpmActionCardService result = cardServiceMap.get(bpmEvent.dataset());
    if (result == null) {
      result = new BpmActionCardService(dbEngine, datasetConfigDao, datasetCardDao, bpmEvent.dataset());
      cardServiceMap.put(bpmEvent.dataset(), result);
    }

    return result;
  }

  private BpmActionListService buildListService(BpmEvent bpmEvent) {
    BpmActionListService result = listServiceMap.get(bpmEvent.dataset());
    if (result == null) {
      result = new BpmActionListService(dbEngine, datasetConfigDao, datasetListDao, bpmEvent.dataset());
      listServiceMap.put(bpmEvent.dataset(), result);
    }

    return result;
  }

  private BpmActionTreeService buildTreeService(BpmEvent bpmEvent) {
    BpmActionTreeService result = treeServiceMap.get(bpmEvent.dataset());
    if (result == null) {
      result = new BpmActionTreeService(dbEngine, datasetConfigDao, datasetTreeDao, bpmEvent.dataset());
      treeServiceMap.put(bpmEvent.dataset(), result);
    }

    return result;
  }

  public AbstractDatasetCardService getCardService(BpmPageContext pageContext) {
    String dataset = pageContext.getPageName() + "." + pageContext.getDataset().getName();
    BpmActionCardService result = cardServiceMap.get(dataset);
    if (result == null) {
      // null就返回默认service
      result = this.defaultActionCardService;
    }

    return result;
  }

  public AbstractDatasetListService getListService(BpmPageContext pageContext) {
    String dataset = pageContext.getPageName() + "." + pageContext.getDataset().getName();
    BpmActionListService result = listServiceMap.get(dataset);
    if (result == null) {
      // null就返回默认service
      result = this.defaultActionListService;
    }

    return result;
  }

  public AbstractDatasetTreeService getTreeService(BpmPageContext pageContext) {
    String dataset = pageContext.getPageName() + "." + pageContext.getDataset().getName();
    BpmActionTreeService result = treeServiceMap.get(dataset);
    if (result == null) {
      // null就返回默认service
      result = this.defaultActionTreeService;
    }

    return result;
  }
}
