package cc.smtweb.system.bpm.spring.dao;

import cc.smtweb.framework.core.SwMap;
import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.db.sqlbuilder.AbstractSelectSqlBuilder;
import cc.smtweb.framework.core.db.sqlbuilder.SqlBuilder;
import cc.smtweb.framework.core.db.sqlbuilder.SqlJoinTable;
import cc.smtweb.system.bpm.core.ui.builder.SqlUtil;
import cc.smtweb.system.bpm.core.ui.entity.dataset.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.List;
import java.util.Map;

@Slf4j
public abstract class AbstractDatasetPlugin {
  @Autowired
  protected DbEngine dbEngine;

  protected AbstractSelectSqlBuilder makeSqlBuilderForTable(BpmDataset dataset, SwMap body, boolean countSql) {
    AbstractSelectSqlBuilder sqlBuilder = createSqlBuilderForTable(dataset, countSql);

    List<Map<String, Object>> filter = (List<Map<String, Object>>)body.get(SqlUtil.PN_FILTER);
    if (filter != null) {
      addSqlWhereForTable(dataset, sqlBuilder, filter, false);
    }

    List<Map<String, Object>> search = (List<Map<String, Object>>) body.get(SqlUtil.PN_SEARCH);

    if (search != null) {
      int filterType = body.readInt(SqlUtil.PN_SEARCH_TYPE, 0);

      addSqlWhereForTable(dataset, sqlBuilder, search, filterType != 0);
    }

    return sqlBuilder;
  }

  protected AbstractSelectSqlBuilder createSqlBuilderForTable(BpmDataset dataset, boolean countSql) {
    AbstractSelectSqlBuilder sqlBuilder = null;

    if (dataset.getDatabase() != null) {
      sqlBuilder = SqlBuilder.createSelect(dataset.getDatabase(), dataset.getTable());
    } else if (dataset.getSql() != null){
      sqlBuilder = SqlBuilder.createDirectSelect(dataset.getSql().getSelect());
    } else {
      return null;
    }

    if (countSql) {
      sqlBuilder.add("count(*) total");

      // 主表统计字段
      addSummaryFields(dataset, sqlBuilder, "");

      // 1:1从表统计字段
      List<BpmDataset> slaves = dataset.getOneToOneDataset();
      if (slaves != null) {
        int index = 0;
        for (BpmDataset slave: slaves) {
          String tableAlias = "b" + (++index);
          if (addSummaryFields(slave, sqlBuilder, tableAlias + ".")) {
            addJoinTable(sqlBuilder, slave, tableAlias);
          }
        }
      }
    } else {
      // 主表字段
      // TODO: 后期不需要返回ID字段，控件设置具体id字段值
      sqlBuilder.add(dataset.readIdFieldName() + " id");

      addFields(dataset, sqlBuilder, "");

      // 1:1从表字段
      List<BpmDataset> slaves = dataset.getOneToOneDataset();
      if (slaves != null) {
        int index = 0;
        for (BpmDataset slave: slaves) {
          String tableAlias = "b" + (++index);
          BpmDataset slaveDataset = slave;
          if (addFields(slaveDataset, sqlBuilder, tableAlias + ".")) {
            addJoinTable(sqlBuilder, slave, tableAlias);
          }
        }
      }
    }
    return sqlBuilder;
  }

  private static SqlJoinTable addJoinTable(AbstractSelectSqlBuilder sqlBuilder, BpmDataset slave, String tableAlias) {
    SqlJoinTable joinTable = sqlBuilder.addJoinTable(slave.getDatabase(), slave.getTable(), tableAlias);
    for (BpmMappingField bpmMappingField : slave.getMasterMapping().getMappingFields()) {
      joinTable.add(bpmMappingField.getMaster(), tableAlias + "." + bpmMappingField.getSlave());
    }

    return joinTable;
  }

  private static boolean addFields(BpmDataset dataset, AbstractSelectSqlBuilder sqlBuilder, String tableAlias) {
    boolean result = false;
    for (Map.Entry<String, BpmField> entry : dataset.getFields().entrySet()) {
      BpmField bpmField = entry.getValue();

//      if (!uiField.isVirtual()) {
        sqlBuilder.add(tableAlias + bpmField.getFieldName() + " " + bpmField.getName());
        result = true;
//      }
    }

    return result;
  }

  private static boolean addSummaryFields(BpmDataset dataset, AbstractSelectSqlBuilder sqlBuilder, String tableAlias) {
    boolean result = false;
    for (Map.Entry<String, BpmField> entry: dataset.getFields().entrySet()) {
      String field = entry.getKey();
      BpmField bpmField = entry.getValue();
      String rule = bpmField.getSummary();

      if (StringUtils.isNotBlank(rule)) {
        String fieldName = tableAlias + bpmField.getFieldName();

        sqlBuilder.add(SqlUtil.makeSummaryField(rule, fieldName) + " " + field);
        result = true;
      }
    }

    return result;
  }

//  protected SelectSqlBuilder makeSqlBuilderForTable(BpmDataset dataset, SwMap query, List<BpmFieldFilter> fieldFilters) {
//    SelectSqlBuilder sqlBuilder = createSqlBuilderForTable(dataset, false);
//
//    List<Map<String, Object>> filters = (List<Map<String, Object>>)query.get(SqlUtil.PN_FILTER);
//
//    if (fieldFilters != null) {
//      for (BpmFieldFilter fieldFilter : fieldFilters) {
//        String keyword = fieldFilter.getKeyword();
//        String fieldName = fieldFilter.getField();
//
//        if (StringUtils.isNotBlank(keyword) && StringUtils.isNotBlank(fieldName)) {
//          // 级联动态数据
//          if (fieldFilter.isCascade()) {
//            if (filters != null) {
//              for (Map<String, Object> filter : filters) {
//                String queryField = (String) filter.get(IBpmDatasetConst.PN_ITEM_FIELD);
//
//                if (fieldName.equals(queryField)) {
//                  String queryKeyword = (String) filter.get(IBpmDatasetConst.PN_ITEM_KEYWORD);
//                  if (StringUtils.isNotBlank(queryKeyword)) {
//                    SqlUtil.addSqlWhere(sqlBuilder, fieldName, fieldFilter.getOp(), queryKeyword);
//                  }
//                  break;
//                }
//              }
//            }
//          } else {
//            SqlUtil.addSqlWhere(sqlBuilder, fieldName, fieldFilter.getOp(), keyword);
//          }
//        }
//      }
//    }
//
//    return sqlBuilder;
//  }

  private void addSqlWhereForTable(BpmDataset dataset, AbstractSelectSqlBuilder sqlBuilder, List<Map<String, Object>> filters, boolean whereOr) {
    if (!filters.isEmpty()) {
      if (whereOr) {
        sqlBuilder.addWhereOrBegin();
      }

      for (Map<String, Object> filter : filters) {
        String keyword = (String) filter.get(IBpmDatasetConst.PN_ITEM_KEYWORD);
        String field = (String) filter.get(IBpmDatasetConst.PN_ITEM_FIELD);
        int op = MapUtils.getIntValue(filter, IBpmDatasetConst.PN_ITEM_OP, 0);

        if (StringUtils.isNotBlank(keyword) && StringUtils.isNotBlank(field)) {
          BpmField bpmField = dataset.findFieldByName(field);

          if (bpmField != null) {
            SqlUtil.addSqlWhere(sqlBuilder, bpmField.getFieldName(), op, keyword);
          } else {
            List<BpmDataset> slaves = dataset.getOneToOneDataset();
            if (slaves != null) {
              int index = 0;
              for (BpmDataset slave : slaves) {
                bpmField = slave.findFieldByName(field);
                if (bpmField != null) {
                  String tableAlias;

                  SqlJoinTable joinTable = sqlBuilder.findJoinTable(slave.getDatabase(), slave.getTable());
                  if (joinTable != null) {
                    // select 重设为 inner join
                    tableAlias = joinTable.getTableAlias();
                  } else {
                    // count 需要加 table
                    tableAlias = "c" + (++index);
                    joinTable = addJoinTable(sqlBuilder, slave, tableAlias);
                  }

                  joinTable.setJoinType(SqlJoinTable.INNER_JOIN);
                  SqlUtil.addSqlWhere(sqlBuilder, tableAlias + "." + bpmField.getFieldName(), op, keyword);
                  break;
                }
              }
            }
          }

          if (bpmField == null) {
            // error
            log.error("not find filter field " + dataset.getName() + "." + field);
          }
        }
      }

      if (whereOr) {
        sqlBuilder.addWhereOrEnd();
      }
    }
  }

  StringBuilder makeLookupSql(BpmFieldLookup fieldLookup, boolean manyToMany) {
    StringBuilder sb = new StringBuilder("SELECT ");

    if (manyToMany) {
      sb.append("m.").append(fieldLookup.getMiddleKeyField()).append(" optKey,")
              .append("b.").append(fieldLookup.getKeyField()).append(" optChildKey,")
              .append("b.").append(fieldLookup.getLabelField()).append(" optValue ").append(" FROM ")
              .append(fieldLookup.getDb()).append(".").append(fieldLookup.getMiddleTable()).append(" m LEFT JOIN ")
              .append(fieldLookup.getDb()).append(".").append(fieldLookup.getTable()).append(" b ON(b.")
              .append(fieldLookup.getKeyField()).append("=m.").append(fieldLookup.getMiddleValueField()).append(") WHERE m.")
              .append(fieldLookup.getMiddleKeyField());
    } else {
      sb.append(fieldLookup.getKeyField()).append(" optKey,").append(fieldLookup.getLabelField()).append(" optValue FROM ")
              .append(fieldLookup.getDb()).append(".").append(fieldLookup.getTable()).append(" WHERE ")
              .append(fieldLookup.getKeyField());
    }

//    StringBuilder sb = new StringBuilder("SELECT ");

    return sb;
  }
}
