package cc.smtweb.system.bpm.web.design.codegen;

import cc.smtweb.framework.core.R;
import cc.smtweb.framework.core.annotation.SwParam;
import cc.smtweb.framework.core.annotation.SwService;
import cc.smtweb.framework.core.db.DbEngine;
import cc.smtweb.framework.core.db.cache.ModelDatabaseCache;
import cc.smtweb.framework.core.db.cache.ModelTableCache;
import cc.smtweb.framework.core.db.vo.ModelDatabase;
import cc.smtweb.framework.core.db.vo.ModelTable;
import cc.smtweb.framework.core.session.UserSession;
import cc.smtweb.framework.file.FilePathGenerator;
import cc.smtweb.system.bpm.core.exception.ModelLoaderError;
import cc.smtweb.system.bpm.core.ui.entity.BpmPage;
import cc.smtweb.system.bpm.core.ui.entity.dataset.BpmDataset;
import cc.smtweb.system.bpm.core.ui.entity.form.BpmForm;
import cc.smtweb.system.bpm.core.ui.entity.form.BpmWidget;
import cc.smtweb.system.bpm.engine.entity.AspModelTreeVO;
import cc.smtweb.system.bpm.engine.ui.codegen.DefaultVelocityEngine;
import cc.smtweb.system.bpm.engine.ui.entity.yaml.YamlDataset;
import cc.smtweb.system.bpm.engine.ui.entity.yaml.YamlPage;
import cc.smtweb.system.bpm.engine.ui.loader.yaml.YamlDatasetEncoder;
import cc.smtweb.system.bpm.engine.ui.loader.yaml.YamlPageEncoder;
import cc.smtweb.system.bpm.spring.BpmConfigBean;
import cc.smtweb.system.bpm.util.CodeGenUtil;
import cc.smtweb.system.bpm.util.YamlUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

// 后台java代码生成
@Slf4j
@SwService
public class JavaCodeService extends AbstractCodeService {
  public static final String SRC_MAIN_JAVA = "/src/main/java/";
  public static final String SRC_MAIN_RES = "/src/main/resources";
  public static final String RESOURCE_PLUGIN = "/plugin/";
  public static final String RESOURCE_STATIC = "/static/bpm/";

  @SwParam
  protected DbEngine dbEngine;

  @SwParam
  private FilePathGenerator filePathGenerator;

  @SwParam
  private BpmConfigBean bpmConfigBean;

  public R info(UserSession us) throws IOException {
    URL url = getClass().getResource("/");

    return R.success(url);
  }

  public R buildEntity(UserSession us, @SwParam("id") Long id) throws IOException {
    R x = checkConfig();
    if (x != null) {
      return x;
    }

    String codeJavaPath = bpmConfigBean.getCodeJavaPath();

    ModelDatabase modelDatabase = ModelDatabaseCache.getInstance().get(id);

    if (modelDatabase == null) {
      return R.error("数据模型对象不存在, id=" + id);
    }

    String codeJavaPackage = bpmConfigBean.getCodeJavaPackage();
    String[] names = codeJavaPackage.split("\\.");

    File entityPath = new File(codeJavaPath + String.join("/", names) + "/entity");

    if (!entityPath.exists()) {
      if (!entityPath.mkdirs()) {
        return R.error("创建目录失败:" + entityPath.getAbsolutePath());
      }
    }

    DefaultVelocityEngine velocityEngine = makeJavaVelocityEngine(codeJavaPackage);

    List <ModelLoaderError> errors = new ArrayList<>();

    for (ModelTable modelTable: ModelTableCache.getInstance().getDbTables(modelDatabase.getId())) {
      File javaFile = new File(entityPath.getAbsolutePath() + "/"
              + CodeGenUtil.underlineToUpperHump(modelTable.getName()) + ".java");

      // 代码生成到临时文件
      velocityEngine.put("table", modelTable);
      R r = velocityEngine.buildUseSign(id, "/codegen/vm/server/entity.java.vm", javaFile);

      if (r != null) {
        errors.add(new ModelLoaderError(0, "生成文件‘" + javaFile.getName() + "’失败，请用tmp文件手动替换！"));
      }
    }

    R r = R.success(entityPath.getAbsolutePath());

    if (!errors.isEmpty()) {
      r.put("errors", errors);
    }

    return r;
  }

  private R checkConfig() {
    if (!bpmConfigBean.isCodeBuild()) {
      return R.error("禁用代码生成(bpm.code-build)");
    }

    String codeJavaPath = bpmConfigBean.getCodeJavaPath();
    if (StringUtils.isBlank(codeJavaPath)) {
      return R.error("未配置Java文件路径(bpm.code-java-path)");
    }

    if (!codeJavaPath.endsWith(SRC_MAIN_JAVA)) {
      return R.error("Java文件路径必须用 '/src/main/java/' 结尾(bpm.code-java-path)");
    }

    File srcPath = new File(codeJavaPath);

    if (!(srcPath.isDirectory() && srcPath.exists())) {
      return R.error("Java文件路径不存在:" + codeJavaPath);
    }

//    String codeJavaResource = bpmConfigBean.getCodeJavaResource();
//    if (StringUtils.isBlank(codeJavaResource)) {
//      return R.error("未配置Java resource文件路径(bpm.code-java-resource)");
//    }
//
//    if (!codeJavaResource.startsWith("/plugin/") || !codeJavaResource.endsWith("/")) {
//      return R.error("Java resource文件路径必须用 '/plugin/' 开头, '微服务名/' 结尾(bpm.code-java-resource)");
//    }

    String packageName = bpmConfigBean.getCodeJavaPackage();
    if (StringUtils.isBlank(packageName)) {
      return R.error("未配置生成Java包名(bpm.code-java-package)");
    }

    if (!packageName.endsWith(".web")) {
      return R.error("包名路径必须用 '.web' 结尾(bpm.code-java-package)");
    }

    String[] names = packageName.split("\\.");

    File webPath = new File(codeJavaPath + String.join("/", names));
    if (!(webPath.isDirectory() && webPath.exists())) {
      return R.error("包目录不存在:" + webPath.getAbsolutePath());
    }

    return null;
  }

  public R buildService(UserSession us, @SwParam("pageId") Long pageId) throws IOException {
    R x = checkConfig();
    if (x != null) {
      return x;
    }

    String modelFullName = dbEngine.queryString("select model_full_name from sw_bpm.bpm_model where model_id=?", pageId);
    if (modelFullName == null) {
      return R.error("页面未提交");
    }

    AspModelTreeVO moduleNode = loadParentPage(pageId);

    if (moduleNode == null) {
      return R.error("未找到页面的模块信息");
    }

    String codeJavaPath = bpmConfigBean.getCodeJavaPath();
    String codeJavaPackage = bpmConfigBean.getCodeJavaPackage();
    JavaPath javaSrcPath = new JavaPath(codeJavaPath, codeJavaPackage);
    String[] packageNames = modelFullName.split("\\.");
    if (packageNames.length < 2) {
      return R.error("未找到页面的节点信息");
    }

    // 生成后台java service代码
    List<String> files = new ArrayList<>();
    BpmPage uiPage = datasetConfigDao.loadPage(pageId);

    DefaultVelocityEngine velocityEngine = new DefaultVelocityEngine();
    for (BpmDataset dataset: uiPage.getDatasets()) {
      String customApi = dataset.getCustomApi();
      if (customApi != null) {
        String modelName;
        switch (dataset.getDataModel()) {
          case BpmDataset.DATA_MODEL_CARD:
            modelName = "Card";
            break;
          case BpmDataset.DATA_MODEL_LIST:
            modelName = "List";
            break;
          case BpmDataset.DATA_MODEL_TREE:
            modelName = "Tree";
            break;
          default:
            return  R.error(dataset.getName() + " 未设置正确的数据模型: " + dataset.getDataModel());
        }

        // Java服务文件路径
        javaSrcPath.setPackages(packageNames);
        // Java服务类名
        String[] customNames = customApi.split("/");
        if (customNames.length > 1) {
          // 配置了路径和类名
          javaSrcPath.setClassName(customNames[customNames.length - 1]);
        } else {
          // 只配置了类名
//          javaSrcPath.setPackages(serviceNames).addPackage(serviceNode.getModelKey());
          javaSrcPath.setClassName(customApi);
        }

        if (!javaSrcPath.makeDirs()) {
          return R.error("创建目录失败:" + javaSrcPath.getAbsolutePath());
        }

        String className = javaSrcPath.getClassName() + "Service";

        File javaFile = new File(javaSrcPath.getAbsolutePath() + "/" + className + ".java");

        if (!javaFile.exists()) {
          velocityEngine.put("package", javaSrcPath.getPackage());
          velocityEngine.put("dataset", dataset);
          velocityEngine.put("className", className);
          velocityEngine.put("resourceUrl", modelFullName + "." + dataset.getName());

          R r = velocityEngine.build("/codegen/vm/server/" + modelName + "Service.java.vm", javaFile);

          if (r != null) {
            return r;
          }

          files.add(javaFile.getAbsolutePath());
        }
      } else {
        // TODO: 附加的Lookup表，通过字段的lookup产生的
      }
    }

    return R.success(files);
  }

  /** 创建配置Yaml文件，项目部署时需要 */
  public R buildYaml(UserSession us, @SwParam("pageId") Long pageId) {
    R x = checkConfig();
    if (x != null) {
      return x;
    }

    String modelFullName = dbEngine.queryString("select model_full_name from sw_bpm.bpm_model where model_id=?", pageId);
    if (modelFullName == null) {
      return R.error("页面未提交");
    }

//    AspModelTreeVO moduleNode = loadParentPage(pageId);
//
//    if (moduleNode == null) {
//      return R.error("未找到页面的模块信息");
//    }

    String codeJavaPath = bpmConfigBean.getCodeJavaPath();
    String[] packageNames = modelFullName.split("\\.");
    if (packageNames.length < 2) {
      return R.error("未找到页面的节点信息");
    }

    // 生成后台java service代码
    List<String> files = new ArrayList<>();
    BpmPage uiPage = datasetConfigDao.loadPage(pageId);

    List<YamlDataset> yamlDatasets = new ArrayList<>();
    for (BpmDataset dataset: uiPage.getDatasets()) {
      yamlDatasets.add(new YamlDatasetEncoder(dataset).encode());
    }

    YamlPage page = new YamlPage();
    page.setId(pageId);
    page.setDestroyOnClose(uiPage.getDestroyOnClose());

    if (!yamlDatasets.isEmpty()) {
      page.setDatasets(yamlDatasets);
    }

    for (BpmForm form: uiPage.getForms()) {
      page.getForms().add(YamlPageEncoder.encode(form));
    }

    // 生成数据集代码
    // 资源文件路径
    String resourceRootPath = codeJavaPath.substring(0, codeJavaPath.length() - SRC_MAIN_JAVA.length()) + SRC_MAIN_RES;
    File yamlFile = new File(resourceRootPath + RESOURCE_PLUGIN + StringUtils.join(packageNames, "/") + ".yaml");
    File resourcePath = yamlFile.getParentFile();
    if (!resourcePath.exists()) {
      if (!resourcePath.mkdirs()) {
        return R.error("创建目录失败:" + resourcePath.getAbsolutePath());
      }
    }

    YamlUtil.writeValue(yamlFile, page);
    files.add(yamlFile.getAbsolutePath());

    return R.success(files);
  }

//  private AspModelTreeVO loadServiceNames(AspModelTreeVO node, Long pageId, List<String> names) {
//    if (pageId.equals(node.getModelId())) {
//      return node;
//    }
//
//    List<AspModelTreeVO> children = node.getChildren();
//
//    if (children != null) {
//      for (AspModelTreeVO child: children) {
//        AspModelTreeVO result = loadServiceNames(child, pageId, names);
//        if (result != null) {
//          names.add(0, node.getModelKey().toLowerCase());
//
//          return result;
//        }
//      }
//    }
//
//    return null;
//  }

  public R buildEvent(UserSession us, @SwParam("pageId") Long pageId) throws IOException {
    R x = checkConfig();
    if (x != null) {
      return x;
    }

    String modelFullName = dbEngine.queryString("select model_full_name from sw_bpm.bpm_model where model_id=?", pageId);
    if (modelFullName == null) {
      return R.error("页面未提交");
    }

    String codeJavaPath = bpmConfigBean.getCodeJavaPath();
    String[] packageNames = modelFullName.split("\\.");
    if (packageNames.length < 2) {
      return R.error("未找到页面的节点信息");
    }

    // 生成后台java service代码
    BpmPage uiPage = datasetConfigDao.loadPage(pageId);

    // 资源文件路径
    String resourceRootPath = codeJavaPath.substring(0, codeJavaPath.length() - SRC_MAIN_JAVA.length()) + SRC_MAIN_RES;
    File eventFile = new File(resourceRootPath + RESOURCE_STATIC + StringUtils.join(packageNames, "/") + "-event.js");

    if (eventFile.exists()) {
      return R.error("事件文件已经存在: " + eventFile.getAbsolutePath());
    }

    File eventDir = eventFile.getParentFile();
    if (!eventDir.exists()) {
      if (!eventDir.mkdirs()) {
        return R.error("创建目录失败:" + eventDir.getAbsolutePath());
      }
    }

    List<String> events = new ArrayList<>();
    for (BpmDataset dataset: uiPage.getDatasets()) {
      events.add(dataset.getName() + "$onLoad");
    }

    for (BpmForm form: uiPage.getForms()) {
      fillEvent(form, events);
    }

    DefaultVelocityEngine velocityEngine = new DefaultVelocityEngine();
    velocityEngine.put("url", modelFullName);
    velocityEngine.put("events", events);

    R r = velocityEngine.build("/codegen/vm/server/event.js.vm", eventFile);

    if (r != null) {
      return r;
    }

    return R.success(eventFile.getAbsolutePath());
  }

  private void fillEvent(BpmWidget widget, List<String> events) {
    Map<String, String> widgetEvents = widget.getEvents();
    if (widgetEvents != null) {
      widgetEvents.forEach((key, value) -> {
        if (StringUtils.isNotBlank(value)) {
          events.add(widget.getName() + "$" + value);
        }
      });
    }

    BpmWidget[] children = widget.getChildren();
    if (children != null) {
      for (BpmWidget child: children) {
        fillEvent(child, events);
      }
    }
  }


  private DefaultVelocityEngine makeJavaVelocityEngine(String codeJavaPackage) {
    return new DefaultVelocityEngine().put("package", codeJavaPackage);
  }
}
