package cc.smtweb.system.bpm.web.design.codegen;

import cc.smtweb.framework.core.R;
import cc.smtweb.framework.core.annotation.SwParam;
import cc.smtweb.framework.core.annotation.SwService;
import cc.smtweb.framework.core.session.UserSession;
import cc.smtweb.framework.file.FilePathGenerator;
import cc.smtweb.system.bpm.core.ui.entity.BpmPage;
import cc.smtweb.system.bpm.engine.entity.AspModelTreeVO;
import cc.smtweb.system.bpm.engine.ui.codegen.DefaultVelocityEngine;
import cc.smtweb.system.bpm.spring.BpmConfigBean;
import cc.smtweb.system.bpm.util.CodeGenUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.tomcat.util.http.fileupload.IOUtils;

import java.io.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

// 前端界面代码生成
@Slf4j
@SwService
public class VueCodeService extends AbstractCodeService {
  @SwParam
  private FilePathGenerator filePathGenerator;

  @SwParam
  private BpmConfigBean bpmConfigBean;

  public R info(UserSession us) throws IOException {
    URL url = getClass().getResource("/");

    return R.success(url);
  }

  /** 生成代码 */
  public R build(UserSession us, @SwParam("pageId") Long pageId, @SwParam("type") String type) throws IOException {
    if (!bpmConfigBean.isCodeBuild()) {
      return R.error("禁用代码生成(bpm.code-build)");
    }

    String codeVuePath = bpmConfigBean.getCodeVuePath();
    if (StringUtils.isBlank(codeVuePath)) {
      return R.error("未配置生成VUE文件路径(bpm.code-vue-path)");
    }

    String codeVueDir = getDir();

    if (!codeVuePath.endsWith("/src/")) {
      return R.error("VUE文件路径必须用 '/src/' 结尾(bpm.code-vue-path)");
    }

    File srcPath = new File(codeVuePath);

    if (!(srcPath.isDirectory() && srcPath.exists())) {
      return R.error("VUE文件路径不存在:" + codeVuePath);
    }

    if ("router".equals(type)) {
      return makeRouterCode(pageId, codeVuePath);
    } else if (dbEngine.queryLong("select model_id from sw_bpm.bpm_model where model_id=?", pageId) == null) {
      return R.error("页面未提交");
    }

    return makePageCode(pageId, codeVuePath, codeVueDir);
  }

  private R makeRouterCode(Long pageId, String codeVuePath) throws IOException {
    AspModelTreeVO moduleNode = loadParentPage(pageId);

    if (moduleNode == null) {
      return R.error("未找到页面的模块信息");
    }

    // src/router/modules/{moduleNode.name}.ts
    List<RouterVO> routers = new ArrayList<>();
    R r = buildRouter(moduleNode, routers, new ArrayList<>());
    if (r != null) {
      return r;
    }

    DefaultVelocityEngine velocityEngine = new DefaultVelocityEngine();
    velocityEngine.put("routers", routers);

    // 生成vue主页面
    File vmFile = new File(codeVuePath + "/router/modules/" + moduleNode.getModelKey() + ".ts");
    r = velocityEngine.buildUseSign(moduleNode.getModelId(), "/codegen/vm/router.ts.vm", vmFile);
    if (r != null) {
      return r;
    }

    return R.success(vmFile.getAbsolutePath());
  }

  private R buildRouter(AspModelTreeVO node, List<RouterVO> routers, List<String> names) {
    if (StringUtils.isBlank(node.getModelKey())) {
      return R.error("页面 '" + node.getModelName() + "' 未设置命名");
    }
    names.add(node.getModelKey());

    List<AspModelTreeVO> children = node.getChildren();

    if (children != null) {
      for (AspModelTreeVO child: children) {
        R r = buildRouter(child, routers, names);
        if (r != null) {
          return r;
        }
      }
    }

    // 最后一级才生成页面路由
    if (node.getChildren() == null) {
      StringBuilder sb = new StringBuilder();
      names.forEach((name) -> sb.append(CodeGenUtil.toLowerHump(name)));
      sb.setCharAt(0, Character.toLowerCase(sb.charAt(0)));

      RouterVO router = new RouterVO();
      router.setFilePath(String.join("/", names));
      router.setPath(String.join("/", names));
      router.setName(getRouterName(names));
      router.setTitle(node.getModelName());

      routers.add(router);
    }

    names.remove(names.size() - 1);

    return null;
  }

  private static String getRouterName(List<String> names) {
    StringBuilder sb = new StringBuilder();
    for (String name: names) {
      sb.append(CodeGenUtil.toUpperHump(name));
    }

    return sb.toString();
  }

  private R makePageCode(@SwParam("pageId") Long pageId, String codeVuePath, String codeVueDir) throws IOException {
    // 重新读取页面，获取加载错误值
    BpmPage uiPage = datasetConfigDao.loadPage(pageId);

    // 创建页面目录
    List<String> names = datasetConfigDao.loadPagePath(uiPage, false);

    File viewPath = new File(codeVuePath + codeVueDir + "/" + String.join("/", names));

    if (!viewPath.exists()) {
      if (!viewPath.mkdirs()) {
        return R.error("创建目录失败:" + viewPath.getAbsolutePath());
      }
    }

    /*
     * 启用VM引擎生成页面
     */
    DefaultVelocityEngine velocityEngine = makeVueVelocityEngine(codeVueDir, uiPage, names);

    // 生成vue主页面
    File vmFile = new File(viewPath.getAbsolutePath() + "/index.vue");
    R r = velocityEngine.buildUseSign(uiPage.getId(), "/codegen/vm/index.vue.vm", vmFile);
    if (r != null) {
      return r;
    }

    // 生成dataset.ts文件
    File tsFile = new File(viewPath.getAbsolutePath() + "/dataset.ts");
    r = velocityEngine.buildUseSign(uiPage.getId(), "/codegen/vm/dataset.ts.vm", tsFile);
    if (r != null) {
      return r;
    }

    // 生成 events.ts 事件文件
    r = velocityEngine.build("/codegen/vm/events.ts.vm", new File(viewPath.getAbsolutePath() + "/event.ts"));
    if (r != null) {
      return r;
    }

    // 复制基础文件
    File commonPath = new File(codeVuePath + "/" + codeVueDir + "/page");
    if (!commonPath.exists()) {
      if (!commonPath.mkdirs()) {
        return R.error("创建目录失败:" + commonPath.getAbsolutePath());
      }
    }

    copy("dataset.ts", commonPath);
    copy("events.ts", commonPath);
    copy("page.types.ts", commonPath);

    return R.success(vmFile.getAbsolutePath());
  }

  private DefaultVelocityEngine makeVueVelocityEngine(String codeVueDir, BpmPage uiPage, List<String> routerNames) {
    DefaultVelocityEngine velocityEngine = new DefaultVelocityEngine();

    UiPageRefLoader pageRefLoader = new UiPageRefLoader();

    velocityEngine.put("page", uiPage).put("pageRef", pageRefLoader.loadPageRef(uiPage, codeVueDir, pageRefLoader.getRouterName(routerNames)));

    return velocityEngine;
  }

  private void copy(String srcFileName, File commonPath) throws IOException {
    File destFile = new File(commonPath.getAbsoluteFile() + "/" + srcFileName);

    if (!destFile.exists()) {
      try (InputStream is = this.getClass().getResourceAsStream("/codegen/ts/" + srcFileName);
           OutputStream os = new FileOutputStream(destFile)) {
        IOUtils.copy(is, os);
      }
    }
  }

  private String getDir() {
    String codeVueDir = bpmConfigBean.getCodeVueDir();
    if (StringUtils.isBlank(codeVueDir)) {
      codeVueDir = "views";
    }
    return codeVueDir;
  }
}
